import typing
from typing import Literal

from padmy.db import Database
from .network import convert_db
from ..logs import logs

try:
    import dash_cytoscape as cyto
except ImportError:
    logs.error("Both dash and dash_cytoscape need to be installed")
    raise

from dash import Dash, Output, Input, html
import json
import networkx as nx

cyto.load_extra_layouts()


class NodeData(typing.TypedDict):
    id: str
    count: int
    label: str


class EdgeData(typing.TypedDict):
    id: str
    source: str
    target: str


class DirectedElement(typing.TypedDict):
    data: NodeData | EdgeData


def get_directed_edges(g: nx.DiGraph) -> list[DirectedElement]:
    directed_edges: list[DirectedElement] = [
        {"data": {"id": f"{src}{tgt}", "source": src, "target": tgt}} for src, tgt in g.edges
    ]
    node_elems: list[DirectedElement] = [{"data": {"id": k, **v}} for k, v in g.nodes.data()]  # type: ignore
    return node_elems + directed_edges


_CIRCULAR_NODE_STYLE = {
    "source-arrow-color": "red",
    "source-arrow-shape": "triangle",
    "line-color": "red",
}

_CIRCULAR_EDGE_STYLE = {
    "source-arrow-color": "red",
    "source-arrow-shape": "triangle",
    "line-color": "red",
}


def get_cycles_styles(g: nx.DiGraph):
    styles = []
    for cycles in nx.simple_cycles(g):
        # Self referencing
        if len(cycles) == 1:
            cycle = cycles[0]
            styles += [
                {
                    "selector": f"#{cycle}",
                    "style": {**_CIRCULAR_NODE_STYLE, "background-color": "red"},
                },
                {
                    "selector": f"#{cycle}{cycle}",
                    "style": {**_CIRCULAR_EDGE_STYLE, "background-color": "red"},
                },
            ]
        else:
            _src, _tgt = cycles
            styles += [
                {
                    "selector": f"#{_src}",
                    "style": {**_CIRCULAR_NODE_STYLE, "background-color": "red"},
                },
                {
                    "selector": f"#{_tgt}",
                    "style": {**_CIRCULAR_NODE_STYLE, "background-color": "red"},
                },
                {"selector": f"#{_src}{_tgt}", "style": _CIRCULAR_EDGE_STYLE},
                {"selector": f"#{_tgt}{_src}", "style": _CIRCULAR_EDGE_STYLE},
            ]
    return styles


# See https://dash.plotly.com/cytoscape/layout for a full list
Layout = Literal["cose", "breadthfirst", "circle", "preset", "random", "grid", "concentric"]
# needs cyto.load_extra_layouts() to be loaded
ExternalLayout = Literal["cose-bilkent", "cola", "euler", "spread", "dagre", "klay"]


def get_layout(
    g: nx.DiGraph,
    *,
    layout: Layout | ExternalLayout = "klay",
    # Eg: {'width': '100%', 'height': '400px'}
    style: dict | None = None,
) -> cyto.Cytoscape:
    elements = get_directed_edges(g)
    _style = style or {"width": "100%", "height": "800px"}
    stylesheet = [
        {"selector": "node", "style": {"label": "data(label)"}},
        {
            "selector": "edge",
            "style": {
                "curve-style": "bezier",
                "source-arrow-shape": "triangle",
            },
        },
        *get_cycles_styles(g),
    ]
    return cyto.Cytoscape(layout={"name": layout}, style=_style, elements=elements, stylesheet=stylesheet)


def run_simple_app(db: Database, port: int = 5555):
    g = convert_db(db)
    cyto_layout = get_layout(g)
    layout_id = "graph-layout"
    cyto_layout.id = layout_id  # type: ignore

    app = Dash(__name__)

    app.layout = html.Div(
        [
            cyto_layout,
            html.Pre(id="cytoscape-tapNodeData-output"),
            html.Pre(id="cytoscape-tapEdgeData-output"),
        ]
    )

    @app.callback(
        Output("cytoscape-tapNodeData-output", "children"),
        Input(layout_id, "tapNodeData"),
    )
    def _on_press_node(data):
        return json.dumps(data, indent=2)

    @app.callback(
        Output("cytoscape-tapEdgeData-output", "children"),
        Input(layout_id, "tapEdgeData"),
    )
    def _on_press_edge(data):
        return json.dumps(data, indent=2)

    app.run(port=port)  # type: ignore



import os
from pathlib import Path


def create_sql_file(sql_dir: Path, position: int):
    """
    Utility to insert a new sql file between other sql files.
    For instance, let's say we have:

        0000_file1.sql
        0001_file2.sql
        0002_file3.sql
        0003_file4.sql
        0004_file5.sql

    and we want to insert a new file between [bold]0002_file2.sql[/bold] and [bold]0003_file3.sql[/bold],
    by specifying a [bold]position[/bold] of 3 we will have:

        0000_file1.sql
        0001_file2.sql
        0002_file3.sql
        0003_new_file.sql
        0004_file4.sql
        0005_file5.sql
    """
    is_before = True
    sql_files = sorted(sql_dir.glob("*.sql"))
    for _file in sql_files:
        _curr_pos, *_file_name = _file.name.split("_")
        _curr_pos, _file_name = int(_curr_pos), "_".join(_file_name)
        if _curr_pos == position:
            is_before = False
        if not is_before:
            _file.rename(f"{sql_dir}/{_curr_pos + 1:04}_{_file_name}")

    os.close(os.open(f"{sql_dir}/{position:04}_new_file.sql", os.O_CREAT))



import contextlib
import tempfile
from math import floor
from typing import cast, Literal

import asyncpg
from rich.progress import Progress
from tracktolib.pg import insert_many, iterate_pg

from ..db import Database, Table, FKConstraint
from ..logs import logs
from ..utils import (
    check_cmd,
    pg_restore,
    pg_dump,
    create_db,
    drop_db,
    exec_psql,
    check_extension_exists,
    parse_pg_uri,
    temp_env,
)


def copy_database(
    from_pg_uri: str,
    from_db: str,
    to_pg_uri: str,
    to_db: str,
    schemas: list[str],
    drop_public: bool = False,
):
    """
    Dumps and recreate the schemas from `from_db` into `to_db`.
    You can optionally specify `drop_public` to drop the public schema
    before restoring and avoid ' ERROR:  schema "public" already exists'
    (useful if you have *public* specified in your schemas list)
    """
    pg_from_infos, pg_target_infos = parse_pg_uri(from_pg_uri), parse_pg_uri(to_pg_uri)
    for cmd in ["pg_dump", "createdb", "dropdb", "pg_restore", "psql"]:
        check_cmd(cmd)

    with tempfile.NamedTemporaryFile(suffix=".dump") as tmp_file:
        # tmp_file = '/tmp/schema.dump'
        with temp_env({"PG_PASSWORD": pg_from_infos["password"]}):
            pg_dump(
                schemas=schemas,
                dump_path=tmp_file.name,
                options=[
                    "-Fc",
                    "--schema-only",
                    "--no-owner",
                    "--no-privileges",
                    "--extension=*",
                ],
                user=pg_from_infos["user"],
                host=pg_from_infos["host"],
                port=pg_from_infos["port"],
                database=from_db,
            )

        _partial_target_infos: dict = dict(pg_target_infos)
        _partial_target_infos.pop("database")
        with temp_env({"PG_PASSWORD": _partial_target_infos.pop("password")}):
            drop_db(
                to_db,
                **_partial_target_infos,
                if_exists=True,
            )
            create_db(database=to_db, **_partial_target_infos)

            if "public" in schemas or drop_public:
                exec_psql(to_db, "DROP SCHEMA public;", **_partial_target_infos)
            pg_restore(
                dump_path=tmp_file.name,
                database=to_db,
                options=["--no-owner", "--no-privileges"],
                **_partial_target_infos,
            )


def get_insert_child_fk_data_query(table: Table, child_table: Table) -> str:
    """
    Gets the query to insert data from a table to the temporary table.
    """
    joins = []

    def _fk_join(tmp_name: str, fk: FKConstraint, fk_index: int) -> str:
        return f"inner join {tmp_name} _s{fk_index} on " + " and ".join(
            f"_s{fk_index}.{column_name} = t.{foreign_column_name}"
            for column_name, foreign_column_name in zip(fk.column_names, fk.foreign_column_names)
        )

    for i, _fk in enumerate(child_table.foreign_keys):
        if _fk.foreign_full_name != table.full_name:
            continue
        joins.append(_fk_join(child_table.tmp_name, _fk, i))

    query = (
        f"""
    INSERT INTO {table.tmp_name} ({table.values})
    SELECT {table.get_values('t')} from {table.full_name} t
    """
        + "\n".join(joins)
        + "\n ON CONFLICT DO NOTHING"
    )

    return query


def get_insert_data_query(table: Table):
    where_str = "and ".join(f"t2.{_pk.column_name} = t1.{_pk.column_name}" for _pk in table.primary_keys)
    query = f"""
    INSERT into {table.tmp_name} ({table.values})
    SELECT {table.values} from {table.full_name} t1
    """
    if where_str:
        query = f"{query}\nwhere not exists(select null from {table.tmp_name} t2 where {where_str})"
    query = f"{query}\nLIMIT $1"
    return query


async def _insert_leaf_table(conn: asyncpg.Connection, table: Table, table_size: int):
    query = f"CREATE TEMP TABLE {table.tmp_name} ON COMMIT DROP AS SELECT * from {table.full_name}"
    args = []
    # if table_size > 0:
    query = f"{query} TABLESAMPLE SYSTEM_ROWS($1)"
    args.append(table_size)
    logs.debug(f"{query} {args}")
    await conn.execute(query, *args)


async def _insert_node_table(conn: asyncpg.Connection, table: Table, table_size: int):
    # Creating the table
    query = f"CREATE TEMP TABLE {table.tmp_name} (LIKE {table.full_name} INCLUDING ALL) ON COMMIT DROP"
    logs.debug(f"Create node table query: {query}")
    await conn.execute(query)

    # Inserting data from child table
    for _child_table in table.child_tables_safe:
        query = get_insert_child_fk_data_query(table, _child_table)
        logs.debug(f"Insert child data query: {query}")
        await conn.execute(query)

    count = cast(int | None, await conn.fetchval(f"SELECT count(*) from {table.tmp_name}"))
    if count is None:
        raise NotImplementedError("Got empty table")

    logs.debug(f"Table {table.tmp_name!r} has {count} rows, expected {table_size}")
    match count:
        case count if count == table_size:
            pass
        case count if count < table_size:
            query = get_insert_data_query(table)
            _limit = table_size - count
            logs.debug(f"Got {count} < {table_size}, inserting data: {query.replace('$1', str(_limit))}")
            await conn.execute(query, _limit)
        case count if count > table_size:
            logs.warning(f"Sample size cannot be reached (got {count}, expected {table_size})")
        case _:
            raise NotImplementedError(f"Got invalid count: {count} (table_size: {table_size})")


async def process_table(table: Table, conn: asyncpg.Connection) -> set[Table]:
    """ """
    if table.has_been_processed:
        raise ValueError(f"Table {table.full_name!r} has already been processed")

    if table.sample_size is None:
        raise ValueError(f"Got empty sample_size for {table.full_name!r}")

    table_size = floor(table.count * table.sample_size / 100)
    logs.debug(
        f"Processing table: {table.full_name} (is_leaf: {table.is_leaf}, "
        f"sample_size: {table.sample_size}, "
        f"table.count: {table.count}, "
        f"table_size: {table_size}"
        ")"
    )
    if table.is_leaf:
        logs.debug(f"\tInserting leaf table {table.full_name}")
        await _insert_leaf_table(conn, table, table_size)
    else:
        if table.children_has_been_processed:
            logs.debug(f"\tInserting node table {table.full_name}")
            await _insert_node_table(conn, table, table_size)
        else:
            logs.debug("\tWaiting for children to be processed")
            return {_child_table for _child_table in table.child_tables_safe if not _child_table.has_been_processed}

    table.has_been_processed = True

    return {_parent_table for _parent_table in table.parent_tables_safe if not _parent_table.has_been_processed}


@contextlib.asynccontextmanager
async def disable_trigger(conn: asyncpg.Connection, *, active: bool = True):
    if not active:
        yield
        return

    await conn.execute("SET session_replication_role = 'replica'")
    try:
        yield
    finally:
        await conn.execute("SET session_replication_role = 'origin'")


async def create_temp_tables(
    conn: asyncpg.Connection,
    tables: list[Table],
    *,
    start_from: Literal["node", "leaf"] = "node",
):
    """
    Creates temporary tables with a sample of the original table.
    The default strategy for the sampling is to start from root tables
    and insert a sample of the data from the child tables.
    """

    # We start from the nodes
    _tables = set(
        table for table in tables if (table.is_root if start_from == "node" else table.is_leaf) and not table.ignore
    )

    if not _tables:
        raise NotImplementedError("No node table found")

    while len(_tables) > 0:
        _parent_tables = set()

        for table in _tables:
            if table.has_been_processed:
                continue

            _parent_tables = _parent_tables | await process_table(table, conn)

        # To avoid infinite loop
        if _tables == _parent_tables:
            _not_processed = [_table for _table in tables if not _table.children_has_been_processed]

            for _table in _not_processed:
                logs.error(_table.full_name)
                for c in _table.child_tables:
                    logs.error(f"\t {c.full_name} ({c.has_been_processed})")
            raise ValueError(
                "Cyclic foreign keys detected. "
                f'Possible tables are: {", ".join(x.full_name for x in _not_processed)}. '
                "Run `analyze` with `--show-graphs` to debug."
            )

        _tables = _parent_tables


async def sample_database(
    conn: asyncpg.Connection,
    target_conn: asyncpg.Connection,
    db: Database,
    *,
    show_progress: bool = False,
    chunk_size: int = 5_000,
    # Deactivate triggers on reinserting to new database
    no_trigger: bool = True,
):
    """
    From top table to bottom
    """

    # getting the leaves
    has_ext = await check_extension_exists(conn, "tsm_system_rows")
    if not has_ext:
        await conn.execute("CREATE EXTENSION tsm_system_rows")

    async with conn.transaction():
        logs.info("Creating temporary tables")
        await create_temp_tables(conn, db.tables)

        # Safety check
        _not_processed_tables = [x.full_name for x in db.tables if not x.has_been_processed]
        if _not_processed_tables:
            raise NotImplementedError(
                f"Found {len(_not_processed_tables)} tables that has not been "
                f'processed: {", ".join(_not_processed_tables)}'
            )
        #
        _table_count: dict[str, int] = {}
        if show_progress:
            logs.info("Loading tables count")
            for table in db.tables:
                _count = await conn.fetchval(f"SELECT count(*) from {table.tmp_name}")
                if _count is None:
                    raise ValueError("Got empty count")
                _table_count[table.tmp_name] = _count

        logs.info("Done creating temporary tables, inserting to new database.")
        async with disable_trigger(target_conn, active=no_trigger):
            with Progress(disable=not show_progress) as progress:
                task1 = progress.add_task("[green]Inserting tables....", total=len(db.tables))
                task2 = progress.add_task("[purple]Inserting chunks....") if show_progress else None
                for table in db.tables:
                    logs.debug(f"Inserting to {table.full_name}")
                    if task2 is not None:
                        progress.reset(task2, total=_table_count[table.tmp_name])
                    query = f"SELECT {table.values} from {table.tmp_name}"

                    async for chunk in iterate_pg(conn, query, chunk_size=chunk_size):
                        logs.debug(f"{table.full_name} ({len(chunk)})")
                        await insert_many(
                            target_conn,
                            table.full_name,
                            [dict(x) for x in chunk],
                            on_conflict="ON CONFLICT DO NOTHING",
                            quote_columns=True,
                        )
                        if task2 is not None:
                            progress.update(task2, advance=chunk_size)
                    progress.update(task1, advance=1)



import dataclasses
import datetime as dt
import difflib
import filecmp
import functools
from pathlib import Path

from asyncpg import Connection
from asyncpg.exceptions import UndefinedTableError

from tracktolib.pg import insert_many
from padmy.logs import logs
from padmy.utils import exec_file, pg_dump, exec_psql_file
from .utils import get_files, iter_migration_files, MigrationFile

_GET_LATEST_COMMIT_QUERY = """
SELECT file_ts, file_name
FROM public.migration
ORDER BY applied_at DESC, file_ts DESC
LIMIT 1
"""

# PG_GLOBAL_DUMP_DIR = GLOBAL_TMP / 'matching-pg-backup' / 'dump'
SETUP_FILE = Path(__file__).parent / "db.sql"


class MigrationError(Exception):
    def __init__(self, msg: str, diff: str):
        self.msg = msg
        self.diff = diff


async def migrate_setup(conn: Connection):
    logs.info("Setting up migration tables")
    await exec_file(conn, SETUP_FILE)
    logs.info("Done")


def _verify_migration(
    database: str,
    schemas: list[str],
    migration_id: str,
    up_file: Path,
    down_file: Path,
    dump_dir: Path,
):
    _before_dump, _after_dump = (
        f"{migration_id}-before.sql",
        f"{migration_id}-after.sql",
    )

    # Dump before
    pg_dump(
        database,
        schemas,
        dump_path=str(dump_dir / _before_dump),
        options=["-E", "utf8", "--schema-only"],
    )

    logs.info(f"Applying {up_file.name}")
    exec_psql_file(database, str(up_file))

    logs.info(f"Applying {down_file.name}")
    exec_psql_file(database, str(down_file))

    # Dump after
    pg_dump(
        database,
        schemas,
        dump_path=str(dump_dir / _after_dump),
        options=["-E", "utf8", "--schema-only"],
    )

    _before_dump_file, _after_dump_file = (
        dump_dir / _before_dump,
        dump_dir / _after_dump,
    )

    _no_diff = filecmp.cmp(_before_dump_file, _after_dump_file)

    if not _no_diff:
        diff = difflib.unified_diff(
            _before_dump_file.read_text().split("\n"),
            _after_dump_file.read_text().split("\n"),
            fromfile=_before_dump_file.name,
            tofile=_after_dump_file.name,
        )
        raise MigrationError(f"Difference found for migration: {migration_id}", diff="\n".join(diff))


def migrate_verify(
    database: str,
    schemas: list[str],
    dump_dir: Path,
    migration_folder: Path,
):
    """
    Verifies that the up/down migration is correct
    """
    _pg_dump = functools.partial(pg_dump, database=database, schemas=schemas)

    migration_files = get_files(migration_folder)
    logs.info(f"Found {len(migration_files)} migration files")

    logs.info("Checking...")
    _down_files: list[MigrationFile] = []
    try:
        for i, (_up_file, _down_file) in enumerate(iter_migration_files(migration_files)):
            logs.info(f"Checking migration {i + 1}/{int(len(migration_files) / 2)}")
            _verify_migration(
                database=database,
                schemas=schemas,
                migration_id=_up_file.file_id,
                up_file=_up_file.path,
                down_file=_down_file.path,
                dump_dir=dump_dir,
            )
            # Need to be reapplied for the next tests
            exec_psql_file(database, str(_up_file.path))
            _down_files.append(_down_file)
    except MigrationError as e:
        logs.error(e.msg)
        logs.error(e.diff)
        raise e

    # Restoring to initial state
    logs.info("Restoring to initial state")
    for i, _down_file in enumerate(reversed(_down_files)):
        logs.info(f"Applying {_down_file.path.name} ({len(_down_files) - i}/{len(_down_files)})")
        exec_psql_file(database, str(_down_file.path))
    # else:
    #     if last_down_file:
    #         exec_psql_file(database, str(last_down_file.path))

    logs.info("Everything ok!")


@dataclasses.dataclass
class LatestMigration:
    timestamp: dt.datetime
    file_name: str

    @classmethod
    def load(cls, data: dict):
        return cls(timestamp=data["file_ts"], file_name=data["file_name"])


class NoSetupTableError(Exception):
    pass


async def _get_latest_migration(conn: Connection):
    try:
        resp = await conn.fetchrow(_GET_LATEST_COMMIT_QUERY)
    except UndefinedTableError as e:
        if 'relation "public.migration" does not exist' in str(e):
            raise NoSetupTableError(
                'Could not find table table "public.migration", did you forget to setup the table'
                ' by running "migration setup" ?'
            )
        else:
            raise
    return LatestMigration.load(dict(resp)) if resp else None


async def migrate_up(
    conn: Connection,
    folder: Path,
    *,
    nb_migrations: int = -1,
):
    """Migrate from the latest migration applied available in the database
    to the latest one in the `folder` dir.
    """
    latest_migration = await _get_latest_migration(conn)
    logs.info(f"Latest timestamp: {latest_migration.timestamp}" if latest_migration else "No previous migration found")
    migration_files = get_files(folder)

    migrations_to_apply = [
        _up_file
        for _up_file, _ in iter_migration_files(migration_files)
        if latest_migration is None
        or (_up_file.ts >= latest_migration.timestamp and _up_file.path.name != latest_migration.file_name)
    ]
    if nb_migrations > 0:
        migrations_to_apply = migrations_to_apply[:nb_migrations]

    if not migrations_to_apply:
        logs.info("No migrations to apply")
        return

    logs.info(f"Found {len(migrations_to_apply)} migrations to apply")
    async with conn.transaction():
        for _migration in migrations_to_apply:
            logs.info(f"Running {_migration.path.name}...")
            try:
                await exec_file(conn, _migration.path)
            except Exception as e:
                logs.error(f'Failed to execute migration_id "{_migration.file_id}"')
                raise e
            commit_data = {
                "file_ts": _migration.ts,
                "file_id": _migration.file_id,
                "migration_type": "up",
                "file_name": _migration.path.name,
            }
            await insert_many(conn, "public.migration", [commit_data])

    logs.info("Done!")


async def migrate_down(
    conn: Connection,
    folder: Path,
    *,
    nb_migrations: int = -1,
    # migration_id: str = None
):
    """
    Rollback `nb_migrations` back (default -1 for all available)
    """
    latest_migration = await _get_latest_migration(conn)

    migration_files = get_files(folder, reverse=True)

    rollback_to_apply = [
        _down_file
        for _, _down_file in iter_migration_files(migration_files)
        if latest_migration is None
        or (_down_file.ts <= latest_migration.timestamp and _down_file.path.name != latest_migration.file_name)
    ]

    if nb_migrations > 0:
        rollback_to_apply = rollback_to_apply[:nb_migrations]

    nb_files = len(rollback_to_apply)
    if nb_files == 0:
        logs.info("No rollback files to apply")
        return

    logs.info(f"Found {nb_files} rollback files to apply")
    async with conn.transaction():
        for _rollback in rollback_to_apply:
            logs.info(f"Running {_rollback.path.name}...")
            try:
                await exec_file(conn, _rollback.path)
            except Exception as e:
                logs.error(f'Failed to execute migration_id "{_rollback.file_id}"')
                raise e

            commit_data = {
                "file_ts": _rollback.ts,
                "file_id": _rollback.file_id,
                "migration_type": "down",
                "file_name": _rollback.path.name,
            }
            await insert_many(conn, "public.migration", [commit_data])

    logs.info("Done!")



import textwrap
import time
import uuid
from pathlib import Path

from rich.markup import escape
from rich.prompt import Prompt

from padmy.logs import logs
from .utils import get_files, iter_migration_files
from .config import MigrationConfig
from padmy.env import CONSOLE


def _get_user_email() -> str | None:
    _config = MigrationConfig.load()
    author = Prompt.ask("[blue]Author[/blue]", default=_config.author, console=CONSOLE)
    _config.author = author
    _config.save()
    return author


def _get_last_migration_name(folder: Path) -> str | None:
    """Returns the most recent migration files"""
    files = get_files(reverse=True, folder=folder)
    if not files:
        return None
    up_file, down_file = next(iter_migration_files(files))
    return up_file.path.name


def create_new_migration(folder: Path):
    """
    Creates 2 new files, up and down
    """
    folder.mkdir(exist_ok=True, parents=True)

    _base_name = f"{int(time.time())}-{str(uuid.uuid4())[:8]}"
    CONSOLE.print(f"\nCreating new migration file ([green]{escape(_base_name)}[/green]):\n")

    last_migration = _get_last_migration_name(folder)
    logs.debug(f"Last migration files: {last_migration}")
    author = _get_user_email()
    logs.debug(f"User email: {author}")

    up_file = folder / Path(f"{_base_name}-up.sql")
    down_file = folder / Path(f"{_base_name}-down.sql")

    file_header = textwrap.dedent(
        f"""
    -- Prev-file: {last_migration or ''}
    -- Author: {author or ''}
    """
    ).strip()

    up_file.write_text(file_header)
    down_file.write_text(file_header.replace("-up", "-down"))

    CONSOLE.print("\nNew files created!\n")



import os
from pathlib import Path
from rich.console import Console

PG_DATABASE = os.getenv("PG_DATABASE", "postgres")
PG_HOST = os.getenv("PG_HOST", "localhost")
PG_PORT = int(os.getenv("PG_PORT", "5432"))
PG_USER = os.getenv("PG_USER", "postgres")
PG_PASSWORD = os.getenv("PG_PASSWORD", "postgres")

# Migration
SQL_DIR = os.getenv("SQL_DIR")
MIGRATION_DIR = os.getenv("MIGRATION_DIR")

_PADMY_FOLDER = Path(os.getenv("PADMY_FOLDER", Path.home() / ".padmy"))
PADMY_CONFIG = _PADMY_FOLDER / "config.json"


CONSOLE = Console(markup=True, highlight=False)



from importlib import reload

import asyncpg
import json
import os
import re
import subprocess
import typing
from contextlib import contextmanager
from pathlib import Path
from piou import Option, Derived, Password
from typing import Sequence, AsyncIterator, Callable, TypeVar, cast, Literal
from functools import partial

from padmy import env
from .env import PG_HOST, PG_PORT, PG_USER, PG_DATABASE, PG_PASSWORD
from .logs import logs

PgHost = Option(PG_HOST, "--host", help="PG Host")
PgPort = Option(PG_PORT, "--port", help="PG Port")
PgUser = Option(PG_USER, "--user", help="PG User")
PgPassword = Option(PG_PASSWORD, "-p", help="PG Password")
PgDatabase = Option(PG_DATABASE, "--db", help="PG Database")


def get_pg_root(
    pg_host: str = Option(PG_HOST, "--host", help="PG Host"),
    pg_port: int = Option(PG_PORT, "--port", help="PG Port"),
    pg_user: str = Option(PG_USER, "--user", help="PG User"),
    pg_password: Password = Option(PG_PASSWORD, "-p", help="PG Password"),
):
    return f"postgresql://{pg_user}:{pg_password}@{pg_host}:{pg_port}"


def get_pg_url(pg_url: str = Derived(get_pg_root), pg_database: str = PgDatabase):
    return f"{pg_url}/{pg_database}"


async def get_pg(pg_url: str = Derived(get_pg_url)):
    return await asyncpg.connect(pg_url)


OnStdErrorFn = Callable[[str], None]


def exec_cmd(
    cmd: Sequence[str | int],
    env: dict | None = None,
    *,
    on_stderr: OnStdErrorFn | None = None,
) -> str:
    _cmd = cmd if isinstance(cmd, str) else " ".join(str(x) for x in cmd)
    logs.debug(f"Executing cmd: {_cmd}")
    stdout, stderr = subprocess.Popen(
        _cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env
    ).communicate()
    if stderr:
        _msg = stderr.decode("utf-8")
        if on_stderr is not None:
            on_stderr(_msg)
        else:
            logs.warning(_msg)
    return stdout.decode("utf-8")


def has_cmd(cmd: str) -> bool:
    return exec_cmd(f'command -v "{cmd}"') != ""


class CommandNotFound(Exception):
    def __init__(self, cmd: str):
        super().__init__()
        self.cmd = cmd


_COMMANDS = {}


def check_cmd(cmd: str):
    global _COMMANDS
    path = exec_cmd(f'command -v "{cmd}"')
    if path == "":
        raise CommandNotFound(cmd=cmd)
    _COMMANDS[cmd] = path.strip()


def get_pg_envs():
    reload(env)
    return {
        "PGPASSWORD": env.PG_PASSWORD,
        "PGUSER": env.PG_USER,
        "PGHOST": env.PG_HOST,
        "PGPORT": str(env.PG_PORT),
    }


class PGError(Exception):
    def __init__(self, msg: str, cmd: str | None = None):
        super().__init__(msg)
        self.msg = msg
        self.cmd = cmd
        self.errors = extract_pg_error(msg)

    @property
    def msg_fmt(self):
        return "=== ERROR ====\n\n" + "\n\n=== ERROR ====\n\n".join(self.errors)


def _on_pg_error(msg: str, cmd: list[str] | str):
    if "ERROR" in msg or "FATAL" in msg:
        _cmd = cmd if isinstance(cmd, str) else " ".join(str(x) for x in cmd)
        raise PGError(msg, cmd=_cmd)


def _get_conn_infos(
    user: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
):
    infos = {"-U": user, "-W": password, "-h": host, "-p": port}
    cmd = []
    for k, v in infos.items():
        if v is not None:
            cmd += [k, v]
    return cmd


def pg_dump(
    database: str,
    schemas: list[str],
    dump_path: str,
    options: list[str] | None = None,
    *,
    user: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    on_stderr: OnStdErrorFn | None = None,
    get_env: bool = True,
):
    _schemas = "|".join(x for x in schemas)
    cmd = [
        _COMMANDS["pg_dump"],
        "-n",
        f"'({_schemas})'",
        "-d",
        database,
    ] + _get_conn_infos(user, password, host, port)
    if options:
        cmd += options

    cmd += [">", dump_path]
    _on_stderr = on_stderr or partial(_on_pg_error, cmd=cmd)
    exec_cmd(cmd, env=get_pg_envs() if get_env else None, on_stderr=_on_stderr)


def pg_restore(
    database: str,
    dump_path: str,
    options: list[str] | None = None,
    *,
    user: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    on_stderr: OnStdErrorFn | None = None,
):
    cmd = [_COMMANDS["pg_restore"], "-d", database, dump_path] + _get_conn_infos(user, password, host, port)
    if options is not None:
        cmd += options
    _on_stderr = on_stderr or partial(_on_pg_error, cmd=cmd)
    exec_cmd(cmd, env=get_pg_envs(), on_stderr=_on_stderr)


def create_db(
    database: str,
    *,
    user: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    on_stderr: OnStdErrorFn | None = None,
):
    cmd = [_COMMANDS["createdb"], database] + _get_conn_infos(user, password, host, port)
    _on_stderr = on_stderr or partial(_on_pg_error, cmd=cmd)
    exec_cmd(cmd, env=get_pg_envs(), on_stderr=_on_stderr)


def drop_db(
    database: str,
    *,
    if_exists: bool = True,
    user: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    on_stderr: OnStdErrorFn | None = None,
):
    cmd = [_COMMANDS["dropdb"]] + _get_conn_infos(user, password, host, port)

    if if_exists:
        cmd.append("--if-exists")
    cmd.append(database)
    _on_stderr = on_stderr or partial(_on_pg_error, cmd=cmd)
    exec_cmd(cmd, env=get_pg_envs(), on_stderr=_on_stderr)


def exec_psql_file(
    database: str,
    sql_file: str,
    *,
    user: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    on_stderr: OnStdErrorFn | None = None,
):
    cmd = [_COMMANDS["psql"], "-f", sql_file, "-d", database] + _get_conn_infos(user, password, host, port)
    _on_stderr = on_stderr or partial(_on_pg_error, cmd=cmd)
    exec_cmd(cmd, env=get_pg_envs(), on_stderr=_on_stderr)


def exec_psql(
    database: str,
    query: str,
    *,
    user: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    on_stderr: OnStdErrorFn | None = None,
):
    cmd = [_COMMANDS["psql"], "-c", f"'{query}'", "-d", database] + _get_conn_infos(user, password, host, port)
    _on_stderr = on_stderr or partial(_on_pg_error, cmd=cmd)
    exec_cmd(cmd, env=get_pg_envs(), on_stderr=_on_stderr)


async def exec_file(conn: asyncpg.Connection, file: Path):
    await conn.execute(file.read_text())


async def iterate_pg(
    conn: asyncpg.Connection,
    query: str,
    *args,
    from_offset: int = 0,
    chunk_size: int = 500,
    timeout: int | None = None,
) -> AsyncIterator[list[asyncpg.Record]]:
    async with conn.transaction():
        cur: asyncpg.connection.cursor.Cursor = await conn.cursor(query, *args)
        if from_offset:
            await cur.forward(from_offset, timeout=timeout)
        while data := await cur.fetch(chunk_size, timeout=timeout):
            yield cast(list[asyncpg.Record], data)


_EXT_EXISTS_QUERY = """
SELECT EXISTS(
    SELECT FROM pg_extension WHERE extname = $1
)
"""


async def check_extension_exists(conn: asyncpg.Connection, extension: str) -> bool:
    _exists = await conn.fetchval(_EXT_EXISTS_QUERY, extension)
    if _exists is None:
        raise NotImplementedError()
    return _exists


_TMP_TABLE_EXISTS_QUERY = """
    SELECT EXISTS (
        SELECT 
        FROM   information_schema.tables 
        WHERE  table_schema LIKE 'pg_temp_%'
        AND table_name = $1
    )
"""


async def check_tmp_table_exists(conn: asyncpg.Connection, table: str) -> bool:
    exists = await conn.fetchval(_TMP_TABLE_EXISTS_QUERY, table)
    if exists is None:
        raise NotImplementedError()
    return exists


X = TypeVar("X")


def get_first(*it: X, fn: Callable) -> X | None:
    for v in it:
        if fn(v):
            return v
    return None


async def get_conn(pool: asyncpg.Pool, fn: Callable):
    async with pool.acquire() as conn:
        await fn(conn=conn)


def get_pg_root_from(source: Literal["from", "to"]):
    _source, _source_lower = source.upper(), source.lower()

    def _get_pg_root(
        pg_host: str = Option(
            os.getenv(f"PG_HOST_{_source}", "localhost"),
            f"--host-{_source_lower}",
            help=f"PG Host {_source}",
            arg_name=f"pg_host_{_source}",
        ),
        pg_port: int = Option(
            int(os.getenv(f"PG_PORT_{_source}", 5432)),
            f"--port-{_source_lower}",
            help=f"PG Port {_source}",
            arg_name=f"pg_port_{_source}",
        ),
        pg_user: str = Option(
            os.getenv(f"PG_USER_{_source}", "postgres"),
            f"--user-{_source_lower}",
            help=f"PG User {_source}",
            arg_name=f"pg_user_{_source}",
        ),
        pg_password: Password = Option(
            os.getenv(f"PG_PASSWORD_{_source}", "postgres"),
            f"--password-{_source_lower}",
            help=f"PG Password {_source}",
            arg_name=f"pg_password_{_source}",
        ),
    ):
        url = f"postgresql://{pg_user}:{pg_password}@{pg_host}:{pg_port}"
        pg_clean = url.replace(f":{pg_password}", ":********")
        logs.debug(f"PG url {source}: {pg_clean}")
        return url

    return _get_pg_root


class PGUriInfos(typing.TypedDict):
    user: str
    password: str
    host: str
    port: int
    database: str | None


_PG_URI_REG = re.compile(r"postgresql:\/\/(?P<user>.*):(?P<password>.*)@(?P<host>.*):(?P<port>\d+)(\/(?P<dbname>.*))?")


def parse_pg_uri(uri: str) -> PGUriInfos:
    if match := _PG_URI_REG.match(uri):
        group = match.groupdict()
        return {
            "user": group["user"],
            "password": group["password"],
            "host": group["host"],
            "port": int(group["port"]),
            "database": group["dbname"],
        }
    raise ValueError(f"Invalid PG uri: {uri}")


@contextmanager
def temp_env(new_env: dict):
    """
    Overrides the environment variables with the given ones
    """
    _env = os.environ.copy()
    os.environ.update(new_env)
    try:
        yield
    finally:
        os.environ.update(_env)


@contextmanager
def temp_pg_env(pg_url: str):
    """
    Overrides the PG environment variables with the given pg_url
    """
    _pg_infos = parse_pg_uri(pg_url)
    _env = {
        "PGHOST": _pg_infos["host"],
        "PGPORT": str(_pg_infos["port"]),
        "PGUSER": _pg_infos["user"],
        "PGPASSWORD": _pg_infos["password"],
    }
    if _pg_infos["database"]:
        _env["PGDATABASE"] = _pg_infos["database"]

    with temp_env(_env):
        yield


async def init_connection(conn: asyncpg.Connection):
    await conn.set_type_codec("jsonb", encoder=json.dumps, decoder=json.loads, schema="pg_catalog")
    await conn.set_type_codec("json", encoder=json.dumps, decoder=json.loads, schema="pg_catalog")


_IS_ERROR_REG = re.compile(r".*(?P<error_type>ERROR|FATAL):(?P<msg>.*)")
_IS_INFO_REG = re.compile(r".*NOTICE:(?P<msg>.*)")
_CLEAN_STR_REG = re.compile(r"^E\s+|\s{2,}", flags=re.MULTILINE)


def _clean_str(msg: str) -> str:
    return _CLEAN_STR_REG.sub(" ", msg).strip()


def extract_pg_error(msg: str) -> list[str]:
    """
    Extracts the errors from a pg error message
    """
    errors = []
    current_error = None
    for line in msg.splitlines():
        if _match_error := _IS_ERROR_REG.match(line):
            if current_error is not None:
                errors.append(_clean_str(current_error))
            current_error = f"{_match_error.group('error_type')}: {_match_error.group('msg')}"
        elif _match_info := _IS_INFO_REG.match(line):
            if current_error is not None:
                errors.append(_clean_str(current_error))
            current_error = None
        else:
            if current_error is not None:
                current_error += f"\n{_clean_str(line)}"

    if current_error is not None:
        errors.append(_clean_str(current_error))

    return errors



import os
import tempfile
from pathlib import Path

from asyncpg import Connection
from piou import Option, Derived, CommandGroup, Password

from padmy.env import SQL_DIR, MIGRATION_DIR
from padmy.logs import logs
from padmy.utils import get_pg, PgHost, PgPort, PgUser, PgDatabase, PgPassword

MIGRATION_DESCRIPTION = """
Utilities to handle schema migrations with PostgresSQL.  
If it's the first time running theses scripts please run 
**`poetry run cli migrate setup`** to setup the database's table
"""

migration = CommandGroup("migrate", help="Migration utilities", description=MIGRATION_DESCRIPTION)

MigrationDir = Option(MIGRATION_DIR or ..., "--sql-dir", help="Directory containing the migration files")
SQLDir = Option(SQL_DIR or ..., "--sql-dir", help="Directory containing the table definitions")


@migration.command(cmd="new-sql")
def new_sql_file(
    sql_dir: Path = SQLDir,
    position: int = Option(1, help="Position where to insert the sql file"),
):
    """
    Utility to insert a new sql file between other sql files.
    """
    from .new_sql import create_sql_file

    create_sql_file(sql_dir, position)


@migration.command(cmd="apply-sql")
async def apply_sql_files(
    pg_conn: Connection = Derived(get_pg),
    sql_dir: Path = SQLDir,
):
    """
    Utility to insert a new sql file between other sql files.
    """
    from ..utils import exec_file

    for file in sorted(sql_dir.glob("*.sql")):
        logs.info(f"Applying {file.name!r}")
        await exec_file(pg_conn, file)


@migration.command(cmd="new", help="Creates 2 new files for a migration (up and down)")
def new_migrate_file(migration_folder: Path = MigrationDir):
    """
    Creates 2 new files for a migration (up and down).
    When applying *up.sql* then *down.sql*, the database must be
    in the state it was previously.
    To check if your **down** file works as expected, you can use the `verify` command.
    """
    from .create_files import create_new_migration

    create_new_migration(migration_folder)


@migration.command(cmd="up", help="Migrate database to the new schema")
async def migrate_up_main(
    pg_conn: Connection = Derived(get_pg),
    sql_dir: Path = MigrationDir,
    nb_migrations: int = Option(1, "-n", "--nb-migrations", help="Number of migrations to apply"),
):
    from .migration import migrate_up, NoSetupTableError

    try:
        await migrate_up(pg_conn, folder=sql_dir, nb_migrations=nb_migrations)
    except NoSetupTableError as e:
        logs.critical(e, exc_info=False)


@migration.command(cmd="down", help="Rollback the database to the specified migration")
async def migrate_down_main(
    conn: Connection = Derived(get_pg),
    sql_dir: Path = MigrationDir,
    nb_rollbacks: int = Option(1, "-n", "--nb-rollbacks", help="Number of rollbacks to apply"),
):
    from .migration import migrate_down, NoSetupTableError

    try:
        await migrate_down(conn, folder=sql_dir, nb_migrations=nb_rollbacks)
    except NoSetupTableError as e:
        logs.critical(e, exc_info=False)


@migration.command(cmd="setup", help="Create the tables that will contain the migration metadata")
async def migrate_setup_main(pg_conn: Connection = Derived(get_pg)):
    from .migration import migrate_setup

    await migrate_setup(pg_conn)


@migration.command(cmd="verify", help="Verify that a migration is valid")
def migrate_verify_main(
    db: str = PgDatabase,
    pg_host: str = PgHost,
    pg_port: int = PgPort,
    pg_user: str = PgUser,
    pg_password: Password = PgPassword,
    schemas: list[str] = Option(..., "--schemas", help="Schemas impacted by the migration"),
    sql_dir: Path = MigrationDir,
):
    os.environ["PG_HOST"] = pg_host
    os.environ["PG_PORT"] = str(pg_port)
    os.environ["PG_USER"] = pg_user
    os.environ["PG_PASSWORD"] = pg_password

    from .migration import migrate_verify, MigrationError

    try:
        with tempfile.TemporaryDirectory() as dump_dir:
            migrate_verify(
                database=db,
                migration_folder=sql_dir,
                schemas=schemas,
                dump_dir=Path(dump_dir),
            )
    except MigrationError as e:
        logs.error(e.msg)
        logs.debug(e.diff)



import asyncio
import itertools
import operator
from functools import partial
from typing import Any, Iterator

import asyncpg
from faker import Faker

from padmy.logs import logs
from ..config import Config, ConfigTable, FieldType, AnoFields
from ..db import load_primary_keys, load_columns_type
from ..utils import get_conn, iterate_pg


def get_update_query(table: str, pks: list[str], fields: list[str], field_types: dict):
    _table_keys = pks + fields
    _set_fields = ", ".join(f"{_field} = u2.{_field}" for _field in fields)
    _values = ", ".join(f"${i + 1}::{field_types[k]}" for i, k in enumerate(_table_keys))
    _where = " and ".join(f"u2.{_pk} = u.{_pk}" for _pk in pks)

    query = f"""
    UPDATE {table} as u
    SET
      {_set_fields}
    from (values
      ({_values})
    ) as u2({', '.join(_table_keys)})
    where {_where}
    """
    return query


def _get_fake_value(faker: Faker, field: FieldType, extra_fields: dict | None = None) -> Any:
    _extra_fields = extra_fields or {}
    match field:
        case "EMAIL":
            return faker.email(**_extra_fields)
        case _:
            raise ValueError(f"Got unimplemented field type {field!r}")


def gen_mock_data(faker: Faker, fields: list[AnoFields], size: int) -> Iterator[dict]:
    for _ in range(size):
        yield {v.column: _get_fake_value(faker, v.type, v.extra_args) for v in fields}


def dict_to_tuple(d: dict, fields: list[str]) -> tuple:
    return tuple(d[k] for k in fields)


async def anonymize_table(
    conn: asyncpg.Connection,
    table: ConfigTable,
    pks: list[str],
    faker: Faker,
    *,
    chunk_size: int = 1_000,
):
    if table.fields is None:
        raise ValueError("Fields must not be empty")
    if not pks:
        raise ValueError(f"No PKs found for {table.full_name!r}")

    query = f"SELECT {', '.join(pks)} from {table.schema}.{table.table}"

    fields = [x.column for x in table.fields]
    fields_types = await load_columns_type(conn, table.schema, table.table, pks + fields)
    update_query = get_update_query(table.full_name, pks, fields, fields_types)

    async with conn.transaction():
        async for chunk in iterate_pg(conn, query, chunk_size=chunk_size):
            mock_data = gen_mock_data(faker, fields=table.fields, size=chunk_size)
            new_data = [dict_to_tuple({**c, **m}, pks + fields) for c, m in zip(chunk, mock_data)]
            await conn.executemany(update_query, new_data)


async def anonymize_db(pool: asyncpg.Pool, config: Config, faker: Faker):
    _tables_to_anonymize = [_table for _table in config.tables if _table.has_ano_fields]

    if not _tables_to_anonymize:
        logs.info("No tables found to anonymize in config file")
        return

    async with pool.acquire() as conn:
        pks = await load_primary_keys(conn, list({_table.schema for _table in _tables_to_anonymize}))

    _pks = {
        _table_name: list(_table_pks)
        for _table_name, _table_pks in itertools.groupby(pks, operator.attrgetter("full_name"))
    }

    await asyncio.gather(
        *[
            get_conn(
                pool,
                partial(
                    anonymize_table,
                    table=_table,
                    pks=[x.column_name for x in _pks[_table.full_name]],
                    faker=faker,
                ),
            )
            for _table in _tables_to_anonymize
        ]
    )



from dataclasses import dataclass, field
from pathlib import Path

# else:
#     from typing import Self
from typing import Literal

import yaml

# if sys.version_info.minor < 11 and sys.version_info.major >= 3:

FieldType = Literal["EMAIL"]

SampleType = float | int


def _check_sample_size(sample: SampleType | None):
    if sample is None:
        return
    if sample < 0 or sample > 100:
        raise ValueError(f"Sample must be a value between 0 and 100 (got {sample})")


@dataclass
class AnoFields:
    column: str
    type: FieldType
    extra_args: dict | None = None

    @classmethod
    def load(cls, data: dict):
        if len(data) == 1:
            column = next(iter(data))
            _type = data[column]
            extra_args = None
        else:
            column = data.pop("column")
            _type = data.pop("type")
            extra_args = data if data else None

        return AnoFields(column=column, type=_type, extra_args=extra_args)


@dataclass
class ConfigTable:
    schema: str
    table: str
    sample: SampleType | None = None

    fields: list[AnoFields] = field(default_factory=list)

    ignore: bool = False

    def __post_init__(self):
        _check_sample_size(self.sample)

    @property
    def full_name(self):
        return f"{self.schema}.{self.table}"

    @property
    def has_ano_fields(self):
        return self.fields is not None

    @classmethod
    def load(cls, table: dict):
        _fields_any: dict | list = table.pop("fields", [])
        _fields: list = [_fields_any] if isinstance(_fields_any, dict) else _fields_any

        return cls(**table, fields=[AnoFields.load(_field) for _field in _fields])


@dataclass
class ConfigSchema:
    schema: str
    sample: SampleType | None = None

    def __post_init__(self):
        _check_sample_size(self.sample)

    @classmethod
    def load(cls, v: str | dict):
        if isinstance(v, str):
            return cls(v)
        else:
            return cls(v["name"], v.get("sample"))


@dataclass
class Config:
    # Sample size in percentage
    sample: SampleType | None = None
    schemas: list[ConfigSchema] = field(default_factory=list)
    tables: list[ConfigTable] = field(default_factory=list)

    def __post_init__(self):
        _check_sample_size(self.sample)

    @classmethod
    def load(cls, sample: SampleType, schemas: list[str]):
        _schemas = [ConfigSchema(x) for x in schemas]
        return cls(sample=sample, schemas=_schemas)

    @classmethod
    def load_from_file(cls, path: Path):
        with path.open("r") as f:
            config = yaml.load(f, Loader=yaml.Loader)
        schemas = [ConfigSchema.load(schema) for schema in config.pop("schemas", [])]
        tables = [ConfigTable.load(table) for table in config.pop("tables", [])]

        return cls(**config, schemas=schemas, tables=tables)



import dataclasses
import datetime as dt
from itertools import groupby
from operator import attrgetter
from pathlib import Path

from padmy.utils import exec_cmd


def get_git_email():
    return exec_cmd(["git", "config", "user.email"]).strip()


@dataclasses.dataclass
class MigrationFile:
    ts: dt.datetime
    file_id: str
    file_type: str
    path: Path


def get_files(folder: Path, reverse: bool = False) -> list[MigrationFile]:
    """Returns the migration files in ascending order"""
    files = []
    for file in folder.glob("*.sql"):
        ts, file_id, file_type = file.name.split("-")
        files.append(
            MigrationFile(
                ts=dt.datetime.fromtimestamp(int(ts)),
                file_id=file_id,
                file_type=file_type,
                path=file,
            )
        )
    return sorted(files, key=attrgetter("ts", "file_id"), reverse=reverse)


def iter_migration_files(files: list[MigrationFile]):
    for _file_id, _files_it in groupby(files, lambda x: x.file_id):
        _files: list[MigrationFile] = list(_files_it)

        _up_files = [x for x in _files if x.file_type == "up.sql"]
        _down_files = [x for x in _files if x.file_type == "down.sql"]

        if len(_up_files) != 1:
            raise ValueError(f'Found {len(_up_files)} "up" files')
        if len(_down_files) != 1:
            raise ValueError(f'Found {len(_down_files)} "up" files')

        yield _up_files[0], _down_files[0]
